#ifndef lint
static	char sccsid[] = "@(#)dev_gp1.c 2.1 87/10/28 Copyr 1986 Sun Micro";
#endif

/*
 * Copyright (c) 1986 by Sun Microsystems, Inc.
 */

/*
 * This file contains the control and viewport functions for the
 * DEV_GP1 package. DEV_GP1 is an example of how to use the GP command
 * interface. This code is not supported by Sun Microsystems, Inc.
 *
 * DEV_GP1 is a functional interface to the GP. In general, there is
 * a routine for each (2d and 3d floating point) GP command. The package
 * asks the GP for one or more contiguous command blocks; these blocks
 * are refered to as a buffer. DEV_GP1 tries to fill the buffer as full
 * as possible before sending it to the GP.
 *
 * All routines
 * have the same basic format: A routine receives as parameters the
 * argments to the GP command and checks to see if the command and its
 * arguments will fit in the buffer. If they can, they is are placed into
 * the buffer; otherwise the buffer is flushed (i.e. sent to the GP) and
 * a new buffer is allocated.  Then the command and arguments are
 * placed in it.
 *
 * There are a couple elements of the GP1 structure deserve comment:
 * 	bytes_left	= the number of bytes left in the buffer; note that
 * 			  GP shared memory and commands are talked about
 * 			  in terms of shorts (16 bits) but we keep track of
 * 			  bytes left because sizeof returns bytes.
 * 	list_type	= the DEV_GP1 interface allows the user to specify
 * 			  a vector list (via move and draw calls). DEV_GP1
 * 			  uses this field to determine if a list is already
 * 			  in progress. DEV_GP1 functions othere than move &
 * 			  draw set this field to GP1_NO_LIST which means that
 * 			  any subsequent move or draw calls must set up the
 * 			  list again.
 * 	cur_ptr		= this points to the word (a 16 bit short) in shared
 * 			  memory into which the next command or data is
 * 			  will be written.
 *
 * Here are a few areas within the package that you might want to change:
 * 1) The transformation of vectors and polygons go thru the same flush
 *    procedure as output primitives and as such incur a pw_lock call which
 *    is a system call. pw_lock does not need to be called for transformations
 *    since nothing is written on the screen. If you plan to do a lot of
 *    transformations, you might want to change the package so that a lock is
 *    not aquired for transforms.
 * 2) If your application knows how many vectors or points in a polyline line
 *    it will have, you can improve the stuffing of these points into the
 *    shared memory by writing a routine that takes a list of points rather
 *    than using the current move/draw interface.
 *
 * Common problems when programming the GP:
 * 1) Writing past the end of the command block(s).
 * 2) Detecting that the window has changed; look at the clip id.
 * 3) Be sure to free the blocks you allocate.
 * 4) Of course, be sure you write commands and arguments in the
 *    correct format. A good check for this is to write a routine which will
 *    print out the contests of a buffer before it is posted to the GP. If
 *    you can read what you have written, chances are the GP can to.
 * 5) Debugging the GP can be tricky. Be aware that even though your program
 *    may be stopped in the debugger (and thus not accessing the GP) other
 *    programs, such as clocktool and perfmeter, will be accessing it. I
 *    suggest that you do not run programs like clocktool while debugging.
 *    Also, if you are running the debugger on the same screen as the
 *    program you are debugging then the debugger will also be using the
 *    GP. This is not such a bad problem; however if you get really stuck,
 *    try running the debugger from a terminal.
 *
 *
 * -rlh, smi -- 9/12/86
 *
 * massive re-write, -eaf 9/2/87
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include <sun/fbio.h>
#include "dev_gp1_internal.h"

static	float	identity[4][4] = {		/* 3D identity matrix */
    {   1.0,	0.0,	0.0,	0.0},
    {   0.0,	1.0,	0.0,	0.0},
    {   0.0,	0.0,	1.0,	0.0},
    {   0.0,	0.0,	0.0,	1.0}
};



extern	char	*malloc() ;
extern	int	gp1_alloc() ;
extern	int	gp1_post() ;

static	void	reset_to_defaults();
	void	_dev_gp1_update_clip_list();


/* Debugging Control */

/*******
#undef  DBG_GP1
#define DBG_GP1(a)  a

#define DO_DBG_BIT_VEC	1
******/

#define	dev_gp1_window_changed(gp1)	\
	((gp1)->pw && (gp1)->clipid != (gp1)->pw->pw_clipdata->pwcd_clipid)


#ifdef DO_DBG_BIT_VEC
#define	DBG_BIT_VEC(x)	x
#else
#define	DBG_BIT_VEC(x)
#endif


/*
 * Allocate a gp1_handle and get a static block.
 * Initialize everything...
 */


GP1_handle
dev_gp1_int_open(pr, n_buffers, error_num)
	struct pixrect	*pr;		/* pixrect  */
	int	n_buffers;	/* number of buffers to allocate at once. */
	int	*error_num;	/* if an error, return the error number */
{
register GP1	*gp1;
    	int	ioctl_fd;	/* fd for ioctls */
	int	static_block;
	int	texture_block;
	struct { unsigned char maj,minr,ser,flags ; } version ;
	struct fbtype	fbtype ;

    if (n_buffers <= 0)
	n_buffers = GP1_DEFAULT_NUMBER_OF_BUFFERS;

    if (pr->pr_ops->pro_rop != gp1_rop) {
	*error_num = DEV_GP1_ERR_NO_GP1;
	return((GP1 *)NULL);
    }

    ioctl_fd = gp1_d(pr)->ioctl_fd;

    if ((static_block = gp1_get_static_block(ioctl_fd)) < 0) {
	*error_num = DEV_GP1_ERR_NO_CONTEXT_BLK;
	return((GP1 *)NULL);
    }


    gp1 = (GP1 *) malloc(sizeof(GP1));
    if (gp1 == (GP1 *)NULL) {
	*error_num = DEV_GP1_ERR_NO_MEMORY;
	return((GP1 *)NULL);
    }

    bzero(gp1, sizeof(GP1)) ;

    gp1->pr = pr ;		/* backwards compatibility */
    gp1->static_block = static_block ;
    gp1->pattern_block = -1 ;
    gp1->minor_devno = gp1_d(pr)->minordev ;
    gp1->fd = ioctl_fd ;
    gp1->fbi = gp1_d(pr)->cg2_index ;
    gp1->shmem = (short *) gp1_d(pr)->gp_shmem ;
    if( ioctl(ioctl_fd, FBIOGTYPE, &fbtype) < 0 )
    {
	perror("dev_gp1: ioctl") ;
	gp1->fbtype = FBTYPE_SUN2GP ;
	gp1->gptype = GP_GP1 ; 
	gp1->ucode_ver = 0 ; 
	gp1->fbdepth = 8 ;
    }
    else
    {
	gp1->fbtype = fbtype.fb_type ;
	if( gp1_ucode_version(pr, &version) == PIX_ERR )
	{
	    fprintf(stderr, "Dev_gp1: unrecognized microcode version\n") ;
	    gp1->gptype = GP_GP1 ;
	    gp1->ucode_ver = 0 ;
	}
	else
	{
	    if( version.flags & 0x10 )
		gp1->gptype = GP_GP2 ;
	    else if( version.flags & 8 )
		gp1->gptype = GP_GPPLUS ;
	    else
		gp1->gptype = GP_GP1 ;
	    gp1->ucode_ver = (version.maj*256 + version.minr)*256 +
			version.ser ;
	}
	gp1->fbdepth = fbtype.fb_depth ;
    }
    gp1->n_buffers = n_buffers ;
    gp1->maxmask = gp1->realmask = gp1->wmask = 1 ;
    gp1->planes = 1 ;
    gp1->list_type = GP1_NO_LIST;
    gp1->clipid = -1;			/* no clip list */
    gp1->cmaplen = gp1->rcmaplen = 2 ;
    gp1->dbctrl = 1 ;

    _gp1_alloc_buffers(gp1,&gp1->out_buffer);
    return (gp1);
}



void
Dev_gp1_reset_to_defaults(gp1)
register GP1	*gp1;
{
register short	*ptr;
	int	i,j ;
static	short	pat[] = {0} ;

    gp1->xmin = gp1->ymin = gp1->zmin = 0.0 ;
    gp1->xmax = gp1->ymax = gp1->zmax = 1.0 ;

    (*gp1->maskfunc)(gp1) ;

    (*gp1->clipfunc)(gp1,1) ;

    MAKE_ROOM(gp1, 6);
    GP1_PUT_SHORT(gp1, GP1_SET_FB_NUM | gp1->fbi) ;
    GP1_PUT_SHORT(gp1, GP1_SET_FB_PLANES | gp1->realmask) ;
    GP1_PUT_SHORT(gp1, GP1_SET_LINE_TEX) ;
    GP1_PUT_SHORT(gp1, 0) ;
    GP1_PUT_SHORT(gp1, 0) ;
    GP1_PUT_SHORT(gp1, 0) ;

    DEV_GP1_SET_CLIP_PLANES(gp1,0) ;
    DEV_GP1_SET_COLOR(gp1, 1);
    DEV_GP1_SET_ROP_MODE(gp1, PIX_SRC);
    DEV_GP1_SET_HIDDEN_SURF(gp1, GP1_NOHIDDENSURF);
    DEV_GP1_SET_LINE_WIDTH(gp1, 1);
    DEV_GP1_SET_TEXTURE_ORG_SCREEN(gp1, 0, 0);
    DEV_GP1_SELECT_MATRIX(gp1, 0);
    Dev_gp1_set_matrix_3d(gp1, 0, identity);
}



void
_gp1_alloc_buffers(gp1,gpbuf)
register GP1	*gp1;
register Gpbuf	*gpbuf ;
{
    static	void	check_bit_vec();

    /*
     * At this point, we should not have any blocks allocated.
     * check_bit_vec will see if some are.
     */
    DBG_BIT_VEC(check_bit_vec());

    while ((gpbuf->bufoffset = gp1_alloc(gp1->shmem,
					 gp1->n_buffers,
					 &(gpbuf->bitvec),
					 gp1->minor_devno,
					 gp1->fd)) == 0)
    	;


    /*
     * subtract 4 words from the total number of bytes in the buffers:
     *  1 word for GP1_USE_CONTEXT
     *  1 word for GP1_EOCL
     *  2 words for the command bit vector
     */
    gpbuf->gpbuffer = gpbuf->cur_ptr = gp1->shmem + gpbuf->bufoffset ;
    gpbuf->words_left = (gp1->n_buffers * GP1_WORDS_PER_BLOCK) - 4 ;
    gpbuf->len = (gp1->n_buffers * GP1_WORDS_PER_BLOCK) ;
    GP1_PUT_SHORT_B(*gpbuf, GP1_USE_CONTEXT | (gp1->static_block & 0x7)) ;
}





#ifdef	DO_DBG_BIT_VEC
/*
 * Look at the bit vector that indicates which blocks may be allocated
 * to a process and determine if it is OK. This routine only works if
 * this process is the only one using the GP1; try not to have pixrect
 * programs running -- no perfmeter or clock -- at lest hide them under
 * another window.
 */


static
void
check_bit_vec(gp1)
GP1	*gp1;
{
    		uint	host;
    		uint	vp;
		char	*msg;
    register	uint	bit_vec;
    register	int	set;

    /*
     * wait until the gp1 is done.
     */
    /* gp1_sync(gp1->shmem, gp1->fd); */

    host  = *((ushort *)gp1->shmem + 3) << 16;		/* hi-order word */
    host |= *((ushort *)gp1->shmem + 4);		/* low-order word */
    vp    = *((ushort *)gp1->shmem + 5) << 16;		/* hi-order word */
    vp   |= *((ushort *)gp1->shmem + 6);		/* low-order word */

    bit_vec = host ^ vp;

    if ((bit_vec & 0x80000000) == 0) {
	msg = "GP1:check_bit_vec -> control block bit == 0.\n";
	goto error;
    }
    if ((bit_vec & 0x0ff) != 0xff) {
	msg = "GP1:check_bit_vec -> static block bits != 0xff.\n";
	goto error;
    }
    /* turn off bits that should be 1 */
    bit_vec &= 0x7fffff00;
    bit_vec &= ~(gp1->cmd_buff_bit_vec);
    if (bit_vec != 0) {
	fprintf(stderr, "GP1:check_bit_vec -> found an allocated block .\n");
	fprintf(stderr, "\tgp1 bit vec:  0x%8.8x\n", gp1->cmd_buff_bit_vec);
	fprintf(stderr, "\tcur bit vec:  0x%8.8x\n", bit_vec);
	return;
    }

    return;

  error:
    fprintf(stderr, msg);
}
#endif	DO_DBG_BIT_VEC




void
Dev_gp1_flush(gp1, free_blocks)
register GP1	*gp1;
 	int	free_blocks;
{
register int	new_clip_id;


    /* Nothing put in blocks, so return... */
    if ((gp1->out_buffer.cur_ptr - 1) <= gp1->out_buffer.gpbuffer) {
	gp1->list_type = GP1_NO_LIST;
	return;
    }


    /* mark end of command list and free blocks - we are guaranteed that
     * there is room for the commands -- see _gp1_alloc_buffers
     */

    if (free_blocks)
    {
	GP1_PUT_SHORT(gp1, GP1_EOCL | GP1_FREEBLKS) ;
	GP1_PUT_INT(gp1, gp1->out_buffer.bitvec) ;
    }
    else
    {
	GP1_PUT_SHORT(gp1, GP1_EOCL) ;
    }


#ifdef	DEBUG
    if(gp1->out_buffer.cur_ptr - gp1->out_buffer.gpbuffer > gp1->out_buffer.len)
    {
	fprintf(stderr, "ALERT: buffer contains %d words!\n",
	    gp1->out_buffer.cur_ptr - gp1->out_buffer.gpbuffer) ;
    }

    if(vpdbg > 0)
    {
	fprintf(stderr,"gp_int_flush: %d words\n",
	    gp1->out_buffer.cur_ptr - gp1->out_buffer.gpbuffer) ;
	gpb_print(gp1->out_buffer.gpbuffer, gp1->out_buffer.cur_ptr) ;
    }
#endif	DEBUG


    DEV_GP1_LOCK(gp1) ;

    (*gp1->clipfunc)(gp1,0) ;
    if( gp1->flags & (GP_RESIZE | GP_CONTAINS_VWP) )
	_dev_gp1_update_clip_list(gp1) ;


    gp1_post(gp1->shmem, gp1->out_buffer.bufoffset, gp1->fd);

    DEV_GP1_UNLOCK(gp1);


    if(free_blocks)
    {
	_gp1_alloc_buffers(gp1,&gp1->out_buffer) ;
    }
    else
    {
	/* re-use the cmd buffers */
	GP1_CLEAR_BUF(gp1) ;
    }
    if(gp1->flags & GP_CONTAINS_VWP)
	dev_gp1_int_viewscale(gp1, &gp1->out_buffer) ;
    gp1->flags = 0 ;
    gp1->list_type = GP1_NO_LIST;
}








/******
 * dev_gp1_int_viewscale(gp,gpbuf)
 *
 * generate scale & xlate factors which will transform the GP clip
 * window ([-1 1],[-1 1],[0 1]) to a subset of the physical screen
 * window determined by gp->gp_view_min/max.  Final Z values only have
 * signifigance in terms of z-buffering.
 *****/

int
dev_gp1_int_viewscale(gp1,gpbuf)
register GP1	*gp1 ;
register Gpbuf	*gpbuf ;
{
	float		x0,y0,z0, x1,y1,z1 ;	/* viewport corners */
	float		xs,ys,zs, xt,yt,zt ;	/* scale & translate for GP */
register short		*ptr ;


	x0 = gp1->xmin * (gp1->win_size.r_width - 1) ;
	y0 = gp1->win_size.r_height - gp1->ymin * (gp1->win_size.r_height - 1) ;
	z0 = gp1->zmin * 65535.0 ;
	x1 = gp1->xmax * (gp1->win_size.r_width - 1) ;
	y1 = gp1->win_size.r_height - gp1->ymax * (gp1->win_size.r_height - 1) ;
	z1 = gp1->zmax * 65535.0 ;

	xs = (x1-x0) * .5 ;
	ys = (y1-y0) * .5 ;
	zs = (z1-z0) ;
	xt = gp1->win_org_x + (x1+x0) * .5 ;
	yt = gp1->win_org_y + (y1+y0) * .5 ;
	zt = z0 ;

	ptr = gpbuf->cur_ptr ;
	*ptr++ = GP1_SET_VWP_3D ;
	GP1_PUT_F(ptr, xs) ;
	GP1_PUT_F(ptr, xt) ;
	GP1_PUT_F(ptr, ys) ;
	GP1_PUT_F(ptr, yt) ;
	GP1_PUT_F(ptr, zs) ;
	GP1_PUT_F(ptr, zt) ;
	GP1_USEUP_B(*gpbuf, 1+6*sizeof(float)/sizeof(short)) ;
}







/*****
 * Dev_gp1_win_change(gp1)
 *
 * obsolete
 *
 *****/

void
Dev_gp1_win_change(gp1)
register	GP1	*gp1 ;
{}














/*******
 * _dev_gp1_update_clip_list(gp1)
 *
 * allocate gp buffer and send new clip list to gp.
 *
 ******/

void
_dev_gp1_update_clip_list(gp1)
register	GP1	*gp1;
{
register short		*ptr ;
register int		i ;
register Rect		*rptr ;
	 Gpbuf		tmp_ptr ;

    /*
     * allocate a block into which we can put the clip list.
     */

    _gp1_alloc_buffers(gp1, &tmp_ptr) ;

    if( gp1->flags & GP_RESIZE )
    {
	ptr = tmp_ptr.cur_ptr ;

	*ptr++ = GP1_SET_CLIP_LIST ;
	*ptr++ = gp1->nclip ;

	for(rptr = gp1->cliplist, i=gp1->nclip; --i >= 0;)
	{
	    *ptr++ = rptr->r_left ;
	    *ptr++ = rptr->r_top ;
	    *ptr++ = rptr->r_width ;
	    *ptr++ = rptr->r_height ;
	    ++rptr ;
	}

	tmp_ptr.cur_ptr = ptr ;
    }

    dev_gp1_int_viewscale(gp1, &tmp_ptr) ;

    GP1_PUT_SHORT_B(tmp_ptr, GP1_EOCL | GP1_FREEBLKS) ;
    GP1_PUT_INT_B(tmp_ptr, tmp_ptr.bitvec) ;

#ifdef	DEBUG
    if(vpdbg > 0)
    {
	fprintf(stderr, "new clip list\n") ;
	gpb_print(tmp_ptr.gpbuffer, tmp_ptr.cur_ptr) ;
    }
#endif	DEBUG

    gp1_post(gp1->shmem, tmp_ptr.bufoffset, gp1->fd ) ;
}
