#ifndef lint
static	char sccsid[] = "%Z%%M% %I% %E% SMI";
#endif



/* This file performs color table management.
 *
 * There are several logical colors:
 *
 * BLACK	black
 * GREY		grey
 * LGREY	light grey
 * WHITE	white
 * RED		bright red
 * GREEN	bright green
 * BLUE		bright blue
 * YELLOW	bright yellow
 * TERRAIN	dark green, exact shading depends on time of day etc.
 * CITY		grey, exact shading depends on time of day, etc.
 * SKY		blue, exact shading depends on time of day, etc.
 * FOG		greyish green, used for terrain far away
 * CLOUD	cloud color
 *
 *
 * How the colors are layed out in the color table depends on how
 * many colors are available.  This depends on whether or not
 * hardware double-buffering is available.
 *
 *
 * If hardware double-buffering is available, then we will use 128 colors
 * for the flight simulator (saving 16 for the window system).  Colors
 * will be assigned as follows:
 *
 * TERRAIN:	32 shades, depending on time of day.
 * FOG:		32 shades, depending on time of day.
 * CITY:	32 shades, depending on time of day.
 * SKY:		32 shades, depending on time of day.
 * BLACK:	1 shade
 * GREY:	1 shade
 * LGREY:	1 shade
 * WHITE:	1 shade
 * RED:		1 shade
 * GREEN:	1 shade
 * BLUE:	1 shade
 * YELLOW:	1 shade
 * CLOUD:	1 shade of grey, depending on time of day.
 *
 * If hardware double-buffering is not available, then we will have to
 * be content with 15 colors, (saving 16 for the window system).  Colors
 * will be assigned as follows:
 *
 * TERRAIN:	3 shades
 * FOG:		same 3 shades as TERRAIN
 * CITY:	3 shades
 * SKY:		1 shade, depending on time of day
 * BLACK:	1 shade
 * GREY:	1 shade
 * LGREY:	1 shade
 * WHITE:	1 shade
 * RED:		1 shade
 * GREEN:	1 shade
 * BLUE:	1 shade
 * YELLOW:	1 shade
 * CLOUD:	1 shade from the CITY set
 *
 * In both cases, BLACK-YELLOW are hardwired constants.
 *
 */


#include <stdio.h>
#include <suntool/sunview.h>
#include <math.h>
#include "parameters.h"
#include "gfx.h"
#include "flight.h"
#include "graphics.h"




/*	CONSTANTS	*/

static	int	Ter_color[] = {128,255,128} ;	/* bright terrain color */
static	int	Fog_color[] = {160,224,160} ;	/* distant terrain color */
static	int	City_color[] = {192,192,192} ;	/* bright city color */
static	int	Sky_color[] = {192,192,255} ;	/* sky blue */
static	int	Cloud_color[] = {192,192,192} ;	/* light grey*/
#ifdef	COMMENT
static	int	Ter_color[] = {64,128,64} ;	/* bright terrain color */
static	int	Fog_color[] = {80,112,80} ;	/* distant terrain color */
static	int	City_color[] = {96,96,96} ;	/* bright city color */
static	int	Sky_color[] = {96,96,128} ;	/* sky blue */
static	int	Cloud_color[] = {96,96,96} ;	/* light grey*/
#endif	COMMENT

static	int	Black_color[] = {0,0,1} ;	/* black */
static	int	Grey_color[] = {96,96,96} ;	/* grey */
static	int	Lgrey_color[] = {192,192,192} ;	/* light grey */
static	int	White_color[] = {255,255,255} ;	/* white */
static	int	Red_color[] = {255,0,0} ;	/* red */
static	int	Green_color[] = {0,255,0} ;	/* green */
static	int	Blue_color[] = {0,0,255} ;	/* blue */
static	int	Yellow_color[] = {255,255,0} ;	/* yellow */



/*	VARIABLES (time of day, visibility)	*/


static	float	Ambient ;	/* ambient light */

static	int	Terrain[2][3] ;	/* terrain adjusted for time of day */
static	int	Fog[2][3] ;	/* distant terrain, adjusted for TOD */
static	int	City[2][3] ;	/* city, adjusted for TOD */
static	int	Sky[2][3] ;	/* sky, adjusted for TOD */
static	int	Cloud[3] ;	/* cloud color, adjusted for TOD */



/*	ACTUAL COLOR TABLE	*/

static	unsigned char	r[256],g[256],b[256] ;



static
load_ramp(number, first,last, color1,color2)
	int	number, first,last, color1[3],color2[3] ;
{
	int	i ;
	int	color_d[3] ;
	int	len ;

	if( first == -1 )
	  first = Color_Mapping[number][0] ;
	else
	  Color_Mapping[number][0] = first ;
	if( last == -1 )
	  last = Color_Mapping[number][1] ;
	else
	  Color_Mapping[number][1] = last ;

	if( first == last )
	{
	  r[first] = color1[0] ;
	  g[first] = color1[1] ;
	  b[first] = color1[2] ;
	}
	else
	{
	  color_d[0] = color2[0] - color1[0] ;
	  color_d[1] = color2[1] - color1[1] ;
	  color_d[2] = color2[2] - color1[2] ;
	  len = last-first ;

	  for(i=first; i<=last; ++i)
	  {
	    r[i] = color1[0] + (i-first)*(color_d[0])/len ;
	    g[i] = color1[1] + (i-first)*(color_d[1])/len ;
	    b[i] = color1[2] + (i-first)*(color_d[2])/len ;
	  }
	}
}



Color_compute(time)
	long	time ;
{
	int		dark[3],bright[3] ;
	struct tm	*tm ;
	float		theta ;

	/* for now, just assume sun moves from East to West, ignore
	   time of year */

	/* our lattitude is 37.5 degrees N */

	tm = localtime(&time) ;
	theta = M_PI*2 * (tm->tm_hour + tm->tm_min*(1./60.))*(1./24.) ;

#ifdef	COMMENT
	Sun_vec.x = .707 ;
	Sun_vec.y = 0.0 ;
	Sun_vec.z = .707 ;
#endif	COMMENT

	Ambient = -cos(theta) ;
	Sun_vec.x = sin(theta) ;
	Sun_vec.y = 0.0 ;
	Sun_vec.z = -cos(theta) ;		/* sin of angle above H */

	if( Ambient > .34 )
	  Day_state = DAY ;
	else if( Ambient > 0.0 )
	  Day_state = DUSK ;
	else if( Ambient > -.14 )	/* sunlight refracts 8 degrees */
	  Day_state = DARK ;
	else
	  Day_state = DARKEST ;


	Ambient = Sun_vec.z + .14 ;
	Ambient += .14 ;
	if( Ambient > 1.0 )
	  Ambient = 1.0 ;
	if( Ambient < 0.0 )
	  Ambient = 0.0 ;

#ifdef	COMMENT
	Day_state = DAY ;
	Ambient = 1.0 ;
#endif	COMMENT


	Terrain[1][0] = Ter_color[0] ;
	Terrain[1][1] = Ter_color[1] ;
	Terrain[1][2] = Ter_color[2] ;
	Terrain[0][0] = Terrain[1][0] * Ambient * .5 ;
	Terrain[0][1] = Terrain[1][1] * Ambient * .5 ;
	Terrain[0][2] = Terrain[1][2] * Ambient * .5 ;

	Fog[1][0] = Fog_color[0] ;
	Fog[1][1] = Fog_color[1] ;
	Fog[1][2] = Fog_color[2] ;
	Fog[0][0] = Fog[1][0] * Ambient * .5 ;
	Fog[0][1] = Fog[1][1] * Ambient * .5 ;
	Fog[0][2] = Fog[1][2] * Ambient * .5 ;

	City[1][0] = City_color[0] ;
	City[1][1] = City_color[1] ;
	City[1][2] = City_color[2] ;
	City[0][0] = City[1][0] * Ambient * .5 ;
	City[0][1] = City[1][1] * Ambient * .5 ;
	City[0][2] = City[1][2] * Ambient * .5 ;

	/* sky:  If visibility > 5 mi, sky is blue.  If visibility == 0,
	   sky is grey.  Interpolate linearly */
	/* Only one shade is supported at this time */

	if( visibility > 5*MILES )
	{
	  Sky[1][0] = Sky_color[0] ;
	  Sky[1][1] = Sky_color[1] ;
	  Sky[1][2] = Sky_color[2] ;
	}
	else
	{
	  float f = visibility/(5*MILES) ;
	  Sky[1][0] = Sky_color[0]*f + Cloud_color[0]*(1-f) ;
	  Sky[1][1] = Sky_color[1]*f + Cloud_color[1]*(1-f) ;
	  Sky[1][2] = Sky_color[2]*f + Cloud_color[2]*(1-f) ;
	}

	Sky[1][0] *= Ambient ;
	Sky[1][1] *= Ambient ;
	Sky[1][2] *= Ambient ;

	Cloud[0] = Cloud_color[0] * Ambient ;
	Cloud[1] = Cloud_color[1] * Ambient ;
	Cloud[2] = Cloud_color[2] * Ambient ;


	if(hwdb)
	{
	  load_ramp(SKY, 8,8, Sky[1], Sky[1]) ;

	  load_ramp(CLOUD, 9,9, Cloud, Cloud) ;

	  load_ramp(TERRAIN, 32,63, Terrain[0],Terrain[1]) ;

	  load_ramp(FOG, 64,95, Fog[0],Fog[1]) ;

	  load_ramp(CITY, 96,126, City[0],City[1]) ;

	  Gfx_putcolormap(Main_gfx, 0,128, r,g,b) ;
	}

	else		/* no hardware D.B. */

	{
	  load_ramp(SKY, 8,8, Sky[1], Sky[1]) ;

	  load_ramp(TERRAIN, 9,11, Terrain[0],Terrain[1]) ;

	  /* distant terrain: Same as TERRAIN */

	  Color_Mapping[FOG][0] = Color_Mapping[TERRAIN][0] ;
	  Color_Mapping[FOG][1] = Color_Mapping[TERRAIN][1] ;

	  load_ramp(CITY, 12,14, City[0],City[1]) ;

	  Color_Mapping[CLOUD][0] = Color_Mapping[CLOUD][1] = 12 ;

	  Gfx_dbuf_putcolormap(Main_gfx, 0,15, r,g,b) ;
	}
}



Color_init()
{
	char	cmsname[CMS_NAMESIZE] ;
	int	dark[3],bright[3] ;
	int	fract ;

	/* for now, set vector to sun = (.26,0.,.97) */

	sprintf(cmsname,"flight%d",getpid()) ;

	if(hwdb)
	{
	  Gfx_sbuf_on(Main_gfx, 128, cmsname) ;
	  r[127] = 255 ; g[127] = 0 ; b[127] = 0 ;
	}
	else		/* no hardware D.B. */
	{
	  Gfx_dbuf_on(Main_gfx, 15, cmsname) ;
	}

	load_ramp(BLACK, BLACK,BLACK, Black_color, Black_color) ;
	load_ramp(WHITE, WHITE,WHITE, White_color, White_color) ;
	load_ramp(LGREY, LGREY,LGREY, Lgrey_color, Lgrey_color) ;
	load_ramp(GREY, GREY,GREY, Grey_color, Grey_color) ;
	load_ramp(RED, RED,RED, Red_color, Red_color) ;
	load_ramp(GREEN, GREEN,GREEN, Green_color, Green_color) ;
	load_ramp(BLUE, BLUE,BLUE, Blue_color, Blue_color) ;
	load_ramp(YELLOW, YELLOW,YELLOW, Yellow_color, Yellow_color) ;
	Color_compute(Now.tv_sec) ;
}




/* handle transition into clouds, fade to cloud_color from regular colors */


Cloud_shade()
{
	int	fract, fract2 ;
	int	terrain[2][3] ;
	int	fog[2][3] ;
	int	city[2][3] ;
	int	i,j ;

	if( Me_Posn.z >= cloud_base_1 )
	  fract = 256 ;
	else if( Me_Posn.z < cloud_base_1 - CLOUD_TRANS )
	  fract = 0 ;
	else
	  fract = (Me_Posn.z - (cloud_base_1-CLOUD_TRANS))/CLOUD_TRANS*256.0 ;

	fract2 = 255-fract ;

	for(i=0; i<2; ++i)
	  for(j=0; j<3; ++j)
	    terrain[i][j] =
	      (Terrain[i][j]*fract2 + Cloud[j]*fract) / 256 ;

	for(i=0; i<2; ++i)
	  for(j=0; j<3; ++j)
	    fog[i][j] =
	      (Fog[i][j]*fract2 + Cloud[j]*fract) / 256 ;

	for(i=0; i<2; ++i)
	  for(j=0; j<3; ++j)
	    city[i][j] =
	      (City[i][j]*fract2 + Cloud[j]*fract) / 256 ;

	load_ramp(TERRAIN, -1,-1, terrain[0],terrain[1]) ;

	load_ramp(CITY, -1,-1, city[0],city[1]) ;


	if(hwdb)
	{
	  load_ramp(FOG, -1,-1, fog[0],fog[1]) ;
	  Gfx_putcolormap(Main_gfx, 0,128, r,g,b) ;
	}

	else		/* no hardware D.B. */
	  Gfx_dbuf_putcolormap(Main_gfx, 0,15, r,g,b) ;
}
