#ifndef lint
static	char sccsid[] = "%Z%%M% %I% %E% SMI";
#endif

/*
 * Copyright (c) 1986 by Sun Microsystems, Inc.
 */

/*
 * this file contains specific control stuff:
 *
 * control_ship()
 *	read control inputs and act accordingly
 *
 *
 *
 *
 *
 * Rules of Aerodynamics:
 *
 *	b = span of wing
 *	c = chord of wing
 *	S = area of wing = b*c
 *	AR = aspect ration = b/c = b^2/S
 *
 *	d = density of the air = .002265 slug/ft^3 = 1.16733 kg/m^3
 *	A = cross section of airflow
 *	V = velocity of airflow
 *	d*A*V = constant		law of continuity
 *	A*V = constant			(d is constant below speed of sound)
 *
 * 	p = static pressure.
 *	dynamic pressure = q = 1/2*d*V^2
 *	bernoulli's equation: p + q = contant
 *
 *	lift = Cl * q * S		Cl varies with angle of attack, etc.
 *					max Cl for no flaps =~ 1.3
 *					max Cl with slotted flaps =~ 2.0
 *
 *
 *	parasitic drag = Dp = Cdp*q*A
 *
 *	induced drag = Di = Cdi*q*S	Cdi varies with angle of attack, etc.
 *	Cdi = k * Cl^2/AR	k >= 1/pi (1/pi = ideal wing)
 *
 *		Cl = lift/(q*S)
 *		Cdi = k * (lift/(q*S))^2/AR
 *		induced drag = k * (lift/(q*S))^2/AR * q * S
 *	induced drag = k * lift^2 / (q*S) / AR
 *
 *	flat plate area = f = Dp/q = Cdp*S
 *		(note that a *real* flat plate has a Cdp = 1.18)
 *
 *	lift/drag = forward velocity / vertical velocity
 *
 *	wing loading = W/S
 *
 *	stall speed: W/S = Clmax * q(stall speed) = Clmax * .5 * d *Vs^2
 *
 *
 * Cdp's from Smith:
 *
 *	plane		Cdp	Flat Plate Area ft^2
 *
 *	C-152		.038	6.14
 *	Tomahawk	.054	6.64
 *	Skipper		.049	6.36
 *	C-172		.036	6.25
 *	Warrior		.034	5.83
 *	Sierra		.034	5.02
 *	Mooney 201	.017	2.81
 *	Arrow		.027	4.64
 *	C-182		.031	5.27
 *	Bonanza		.019	3.47
 *
 * Thus, for C-172:
 *
 *	Cdp = .036
 *	f = .5806 m^2
 *	S = 16.128 m^2
 *	W = 1088 kg


 * (from rec.aviation:)
 * From Kershner's "The Advanced Pilot's Flight Manual",
 * 
 *    Vs=Vs(max)*SQRT(W/Wmax)
 * 
 *   Where Vs is stall speed at current weight
 *         Vs(max) is stall speed at max weight
 *         W is current weight
 *         Wmax is max weight
 * 
 * The factor SQRT(W/Wmax) is often called the performance factor as it is
 * also used in calculating many other weight dependent changes such as
 * performance speeds for climbs and approaches.
 * 
 */


#include <stdio.h>
#include <math.h>
#include "graphics.h"
#include "flight.h"
#include "object_types.h"



extern	int	debug_level ;

#ifdef	DEBUG
#define	register
#endif	DEBUG



/****
 *
 * read control inputs, update ship's position/velocity/"up" vectors
 *
 ****/


#define	grav	9.8		/* force of gravity */

control_ship()
{
	Point3df	new_delta ;
	float		ms_x, ms_y ;

	/* update engine power */

	Me_Power += Dtime*Kp1*(Me_Throttle-Me_Power) ;


	/* Update velocity
	 *
	 * Work this by adding energies:
	 *
	 * the engine adds energy to the plane:  Power*Dtime
	 * drag subtracts power from the plane:  D*dist = D*speed*Dtime
	 * altitude is energy:			 W*delta_alt = W*speed.z*Dtime
	 * kinetic energy:			 Ek = M*speed*speed/2
	 * there are two kinds of drag:
	 *   parasitic:				 Dp = k*speed
	 *   induced:				 Di = ??? (ignore for now)
	 * weight:				 M*g
	 *
	 * from frame to frame:
	 *
	 *   Ek' = Ek - M*g*speed.z*Dtime + Power*Dtime - D*speed*Dtime
	 *
	 *
	 * errors: I'm not sure about parasitic drag.  Induced drag
	 *	probably depends on angle of bank as well as airspeed
	 *
	 */

	{
	  register float Ek ;		/* kinetic energy */
	  float D ;			/* total drag */

	  D = Dind*Me_Speed ;

	  Ek = Me_Mass*Me_Speed*Me_Speed ;
	  Ek -= Me_Mass*grav*Me_Speed*Me_Forward.z*Dtime ;
	  Ek += PowerMax*Me_Power*Dtime ;
	  Ek -= D*Me_Speed*Dtime ;
	  Ek /= Me_Mass ;
	  Me_Speed = sqrt(Ek) ;
	}



	/* read mouse input */


	if(mouse_control)
	{
	  ms_x = VIEWSCALE*(Mouse_ax - .5) ;
	  ms_y = VIEWSCALE*(Mouse_ay - (1.0-VIEWSIZE/2.0)) ;

	  Me_Roll += ms_x * Dtime * ROLL_SCALE ;
	  if(Me_Roll > M_PI) Me_Roll -= (M_PI*2) ;
	  else if(Me_Roll < -M_PI) Me_Roll += (M_PI*2) ;
	  Roll_cos = cos(Me_Roll) ;
	  Roll_sin = sin(Me_Roll) ;

	  Me_Pitch -= ms_y * Dtime * Roll_cos * PITCH_SCALE ;
	  if(Me_Pitch > (M_PI/2))
	  {
	    Me_Roll += M_PI ;
	    Me_Heading += M_PI ;
	    Me_Pitch = M_PI - Me_Pitch ;
	  }
	  else if(Me_Pitch < -(M_PI/2))
	  {
	    Me_Roll += M_PI ;
	    Me_Heading += M_PI ;
	    Me_Pitch = -M_PI - Me_Pitch ;
	  }
	  Pitch_cos = cos(Me_Pitch) ;
	  Pitch_sin = sin(Me_Pitch) ;

	  Me_Heading -= ms_y * Dtime * Roll_sin * PITCH_SCALE ;
	}

	/* lets do rate-of-turn right.  Assume coordinated flight up to
	   a bank of 60 degrees.  If

	    Ac := centripital accel, m/sec**2
	    r := radius of turn, m
	    s := speed, m/sec
	    w := rate of turn, radians/sec.
	    g := force of gravity, m/sec**2
	    roll := roll angle

	    derivation:
	     1) Ac = s**2 / r		(centripital accel in turn)
	     2) Ac = g*tan(roll)	(reqd. for coordinated flt)
	     3) r = s**2/tan(roll)/g	( (1) + (2) )
	     4) w = s/r
	     5) w = g*tan(roll)/s	( (3) + (4) )
	*/


	{
	  register float tan ;
	  if( Roll_cos < .5  &&  Roll_cos > -.5 ) tan = 1.73 ;
	  else tan = Roll_sin/Roll_cos ;

	  Me_Heading += grav*tan/Me_Speed * Dtime ;
	}
	if(Me_Heading > 6.2832) Me_Heading -= 6.2832 ;
	else if(Me_Heading < 0.0) Me_Heading += 6.2832 ;

	Head_cos = cos(Me_Heading) ;
	Head_sin = sin(Me_Heading) ;

	Me_Forward.x = Head_sin * Pitch_cos ;		/* ok */
	Me_Forward.y = Head_cos * Pitch_cos ;
	Me_Forward.z = Pitch_sin ;

	Me_Up.x =  Head_cos*Roll_sin - Head_sin*Pitch_sin*Roll_cos ; /* ok */
	Me_Up.y = -Head_sin*Roll_sin - Head_cos*Pitch_sin*Roll_cos ;
	Me_Up.z = Pitch_cos * Roll_cos ;

	Me_Right.x = Roll_cos*Head_cos + Roll_sin*Pitch_sin*Head_sin ; /* ok */
	Me_Right.y = -Roll_cos*Head_sin + Roll_sin*Pitch_sin*Head_cos ;
	Me_Right.z = -Roll_sin*Pitch_cos ;

	Me_Posn.x += Me_Speed*Dtime*Me_Forward.x ;
	Me_Posn.y += Me_Speed*Dtime*Me_Forward.y ;
	Me_Posn.z += Me_Speed*Dtime*Me_Forward.z ;

	Me_Posn.x += wind_sin * Dtime ;
	Me_Posn.y += wind_cos * Dtime ;

	/* build matrix that maps ship coords to WC */

	Me_Matrix[0][0] = Me_Right.x ;
	Me_Matrix[0][1] = Me_Right.y ;
	Me_Matrix[0][2] = Me_Right.z ;
	Me_Matrix[0][3] = 0.0 ;
	Me_Matrix[1][0] = Me_Up.x ;
	Me_Matrix[1][1] = Me_Up.y ;
	Me_Matrix[1][2] = Me_Up.z ;
	Me_Matrix[1][3] = 0.0 ;
	Me_Matrix[2][0] = Me_Forward.x ;
	Me_Matrix[2][1] = Me_Forward.y ;
	Me_Matrix[2][2] = Me_Forward.z ;
	Me_Matrix[2][3] = 0.0 ;
	Me_Matrix[3][0] = Me_Posn.x ;
	Me_Matrix[3][1] = Me_Posn.y ;
	Me_Matrix[3][2] = Me_Posn.z ;
	Me_Matrix[3][3] = 1.0 ;

	/* build inverse matrix that maps World coords to Ship */
	/* (no need to do conventional invert with special case matrices */

	Me_imatrix[0][0] = Me_Right.x ;
	Me_imatrix[0][1] = Me_Up.x ;
	Me_imatrix[0][2] = Me_Forward.x ;
	Me_imatrix[0][3] = 0.0 ;
	Me_imatrix[1][0] = Me_Right.y ;
	Me_imatrix[1][1] = Me_Up.y ;
	Me_imatrix[1][2] = Me_Forward.y ;
	Me_imatrix[1][3] = 0.0 ;
	Me_imatrix[2][0] = Me_Right.z ;
	Me_imatrix[2][1] = Me_Up.z ;
	Me_imatrix[2][2] = Me_Forward.z ;
	Me_imatrix[2][3] = 0.0 ;
	Me_imatrix[3][0] =  -Me_Posn.x * Me_imatrix[0][0] +
			    -Me_Posn.y * Me_imatrix[1][0] +
			    -Me_Posn.z * Me_imatrix[2][0] ;
	Me_imatrix[3][1] =  -Me_Posn.x * Me_imatrix[0][1] +
			    -Me_Posn.y * Me_imatrix[1][1] +
			    -Me_Posn.z * Me_imatrix[2][1] ;
	Me_imatrix[3][2] =  -Me_Posn.x * Me_imatrix[0][2] +
			    -Me_Posn.y * Me_imatrix[1][2] +
			    -Me_Posn.z * Me_imatrix[2][2] ;
	Me_imatrix[3][3] =  1.0 ;



	/* keep track of where we've been */

	trail_dtime += Dtime ;
	if( trail_dtime > 10. )
	{
	  trail_dat[trail_ptr].x = Me_Posn.x ;
	  trail_dat[trail_ptr].y = Me_Posn.y ;
	  trail_ptr = (trail_ptr+1)%MAX_TRAIL ;
	  ++trail_num ;
	  trail_dtime -= 10. ;
	}
}
