/*	@(#)object_types.h 1.3 89/03/10 SMI	*/


#ifndef	OBJECT_TYPES
#define	OBJECT_TYPES

/*
 * this file is type definitions for object descriptions.
 *
 * Instead of a display "list", the flight simulator uses a display "tree".
 * There are several types of nodes in a display tree.  Most nodes
 * are terminal nodes (polylines, polygons etc.).  Some nodes are
 * processed and proceed linearly to the next node (object instancing etc.)
 * Other nodes split two or more ways (bsp nodes).  Some nodes make
 * a decision and continue one of several ways (view threshold nodes).
 *
 * The overall tree structure is this:
 *
 *   A linked list of Terrain nodes is built in terrain.c.  Each terrain
 *   note contains a pointer to an optional sub-tree.  This sub-tree
 *   is drawn relative to the origin (usually xmin,ymin) corner of
 *   the terrain node.
 *
 *
 * note that in the case of named_object instancing, a node may have
 * more than one parent.
 *
 * These are the currently defined node types:
 */


typedef	enum node_type {
	  NAMED_OBJECT,		/* root of a subtree.  name is filename
				   for auto-loading from disk		*/
	  BSP_NODE,		/* bsp tree node			*/
	  LIST_NODE,		/* list of objects to draw		*/
	  POLYLINE_NODE,	/* linked list of polylines		*/
	  POLYPOINT_NODE,	/* linked list of polypoints		*/
	  STREET_NODE,		/* line by day, points by night		*/
	  POLYGON_NODE,		/* linked list of polygons		*/
	  G_POLYGON_NODE,	/* linked list of gouraud shaded pgons	*/
	  TERRAIN_NODE,		/* terrain mesh data			*/
	  G_TERRAIN_NODE,	/* gouraud shaded terrain mesh data	*/
	  TRANSFORM_NODE,	/* modelling transform			*/
	  VIEW_NODE,		/* view threshold list			*/
	  VASI_NODE,		/* vasi lights				*/
	  BEACON_NODE,		/* Airport Beacon			*/
	} Object_Type ;

/*
 * All nodes contain a node-type field and a pointer to the function that
 * can process the node.  All nodes also contain a client-data field for
 * storing processed data (such as quad meshes that correspond to terrain
 * data)
 * All nodes contain this generic data:
 *	o type field
 *	o pointer to function to process this node:
 *		(*func)(node, viewpt, matrix)
 *			Desc_Node	*node ;
 *			Point3df	*viewpt ;
 *			Matrix3df	*matrix ;
 *
 * Some nodes contain this info:
 *
 *	o client data for storing pre-processed data for speed
 *	o last-reference time for garbage collection of above
 *
 * type-dependent data is as follows:
 *
 * each NAMED_OBJECT node contains the following information:
 *	o name (also used as filename)
 *	o pointer to next object in named object hash table
 *	o object class (NOT_LOADED, SCENERY, etc)
 *	o bounding box for trivial-reject clipping
 *	o pointer to child node
 *	o pointer to class-dependent data.
 *
 * A TRANSFORM_NODE contains the following info:
 *	o transformation matrix
 *	o inverse transform (for viewing)
 *	o pointer to child node
 *
 *
 * each VIEW_NODE contains the following information:
 *	o Point3df center of object
 *	o visibility threshold; if the object is further away than this
 *	  threshold, this view node is used; otherwise, the next view
 *	  node in the list is tested.  Testing continues until a node
 *	  with a small enough visibility threshold is found.  Successive
 *	  nodes contain more detail.
 *	o pointer to next view node in list
 *	o pointer to child node to display if view passes test, NULL
 *	  implies that nothing is visible at this range.
 *
 * This is the information in a BSP_NODE:
 *	o Point3df center of object
 *	o Point3df normal vector
 *	o color of polygon front face (or -1 if front not visible)
 *	o color of polygon back face (or -1 if back not visible)
 *	o color of polygon front trim (or -1 if none)
 *	o color of polygon back trim (or -1 if none)
 *	o # loops in polygon (or 0 if no polygon)
 *	o # vertices in each loop, up to 3 loops
 *	o pointer to vertices
 *	o pointer to front child node if any
 *	o pointer to back child node if any
 *
 * This is the information in a LIST_NODE:
 *	o array of pointers to objects.
 *
 * This is the information in a POLYLINE or POLYPOINT:
 *	o color
 *	o # points
 *	o pointer to array of Point3df points
 *	o pointer to next polyline, if any
 *
 * This is the information in a STREET:
 *	o color for line during day (or -1)
 *	o color for points during night (or -1)
 *	o # points
 *	o pointer to array of Point3df points
 *	o pointer to next STREET, if any
 *
 * This is the information in a POLYGON:
 *	o color
 *	o trim color (or -1 if none)
 *	o # loops in polygon
 *	o # vertices in each loop, up to 3 loops
 *	o pointer to vertices
 *	o pointer to next polygon, if any
 *
 * This is the information in a G_POLYGON:
 *	o # loops in polygon
 *	o # vertices in each loop, up to 3 loops
 *	o pointer to color-vertices
 *	o pointer to next polygon, if any
 *
 * This is the information in a TERRAIN_NODE
 *	o lattitude and longitude (root nodes only)
 *	o x,y origin (row 0, column 0)
 *	o Point3df corners of patch
 *	o base color of terrain (used for shading)
 *	o shading style, 0=flat, 1=gouraud
 *	o x,y size of patch
 *	o # rows, columns of profile data
 *	o delta-x,y between profiles (derived from size and dimension)
 *	o time last displayed
 *	o pointer to array of Terrain_data points
 *	o pointer to subtree, if any
 *	o name of file this data taken from, if any
 *	o maximum view distance.  If further than this distance, only
 *	  a single polygon will be drawn.
 *
 * This is the information in a VASI_NODE
 *	o type of vasi, 0 = 2-bar,
 *	o position of first light
 *	o vector of first light
 *	o position of second light
 *	o vector of second light
 *
 * This is the information in a BEACON_NODE
 *	o type of beacon, 0 = civilian land
 *	o position of light
 *	o light state
 *	o frames remaining before light changes flashes again
 *
 *
 */



#include <sys/types.h>
#include "parameters.h"


#define	MAX_LOOP	3


extern	int	draw_named_object(),
		draw_bsp_node(),
		draw_list_node(),
		draw_polyline_node(),
		draw_polypoint_node(),
		draw_street_node(),
		draw_polygon_node(),
		draw_g_polygon_node(),
		draw_terrain_node(),
		draw_g_terrain_node(),
		draw_transform_node(),
		draw_view_node(),
		draw_vasi_node(),
		draw_beacon_node() ;

typedef	enum object_class {
	  OBJ_STATIC,		/* scenery and things */
	} Object_Class ;



typedef	struct {	/* terrain data type */
	  short	color ;
	  short	elev ;
	} Terrain_data ;


typedef	struct named_object	Named_Object ;
typedef	struct bsp_node		Bsp_Node ;
typedef	struct list_node	List_Node ;
typedef	struct polyline_node	Polyline_Node ;
typedef	struct street_node	Street_Node ;
typedef	struct polygon_node	Polygon_Node ;
typedef	struct g_polygon_node	G_Polygon_Node ;
typedef	struct terrain_node	Terrain_Node ;
typedef	struct transform_node	Transform_Node ;
typedef	struct view_node	View_Node ;
typedef	struct vasi_node	Vasi_Node ;
typedef	struct beacon_node	Beacon_Node ;



typedef struct desc_node {
	  Object_Type	type ;		/* object type			*/
	  int		(*func)() ;	/* pointer to func to draw this	*/
	} Desc_Node ;



struct named_object {
	  Object_Type	type ;		/* = NAMED_OBJECT */
	  int		(*func)() ;	/* = draw_named_object */
	  long int	last_ref ;
	  char		*name ;		/* object's name	*/
	  Named_Object	*next_hash ;	/* next in hash table	*/
	  Object_Class	class ;		/* object type		*/
	  Point3df	b0,b1 ;		/* bounding box		*/
	  Desc_Node	*child ;	/* first child		*/
	  caddr_t	client ;	/* client data		*/
	} ;


struct view_node {
	  Object_Type	type ;		/* = VIEW_NODE */
	  int		(*func)() ;	/* = draw_view_node */
	  Point3df	center ;	/* for computing distance	*/
	  float		thresh ;	/* min. distance to use this descr */
	  View_Node	*next ;		/* ptr to next level of descr	*/
	  Desc_Node	*child ;	/* first item in descriptor	*/
	} ;


struct bsp_node {
	  Object_Type	type ;		/* = BSP_NODE			*/
	  int		(*func)() ;	/* draw_bsp_node()		*/
	  Point3df	center ;	/* center of node		*/
	  Point3df	normal ;	/* normal vector		*/
	  int		fcolor ;	/* front color (-1 => invisible) */
	  int		bcolor ;	/* back color (-1 => invisible)	*/
	  int		ftrim ;		/* front trim color		*/
	  int		btrim ;		/* back trim color		*/
	  int		nloop ;		/* # loops (up to 3)		*/
	  int		loops[MAX_LOOP] ;
	  Point3df	*vertices ;
	  Desc_Node	*front_child, *back_child ;
	} ;

struct list_node {
	  Object_Type	type ;		/* = LIST_NODE			*/
	  int		(*func)() ;	/* draw_list_node()		*/
	  Desc_Node	*children[10];	/* up to ten items in list	*/
	} ;

struct street_node {
	  Object_Type	type ;		/* = STREET_NODE		*/
	  int		(*func)() ;	/* draw_polyline_node()		*/
	  int		linecolor ;
	  int		dotscolor ;
	  int		npts ;
	  Point3df	*points ;
	  Street_Node	*next ;
	} ;

struct polyline_node {
	  Object_Type	type ;		/* = POLYLINE/POINT_NODE	*/
	  int		(*func)() ;	/* draw_polyline_node()		*/
	  int		color ;
	  int		npts ;
	  Point3df	*points ;
	  Polyline_Node	*next ;
	} ;

struct polygon_node {
	  Object_Type	type ;		/* = POLYGON_NODE		*/
	  int		(*func)() ;	/* draw_polygon_node()		*/
	  int		color ;
	  int		trimcolor ;
	  int		nloop ;
	  int		loops[MAX_LOOP] ;
	  Point3df	*vertices ;
	  Polygon_Node	*next ;
	} ;

struct g_polygon_node {
	  Object_Type	type ;		/* = G_POLYGON_NODE		*/
	  int		(*func)() ;	/* draw_g_polygon_node()	*/
	  int		nloop ;
	  int		loops[MAX_LOOP] ;
	  Point3dfc	*vertices ;
	  G_Polygon_Node *next ;
	} ;

struct terrain_node {
	  Object_Type	type ;		/* = TERRAIN_NODE		*/
	  int		(*func)() ;	/* draw_terrain_node()		*/
	  float		lat, lon ;	/* lat & lon of base		*/
	  float		x0, y0 ;	/* base point in M.C.		*/
	  Point3dfc	p0,p1,p2,p3 ;	/* corner points		*/
	  float		max_ht ;	/* highest point in terrain	*/
	  int		color ;		/* base color for shading	*/
	  float		sx,sy ;		/* size				*/
	  int		nx, ny ;	/* # rows, columns of data	*/
	  float		dx, dy ;	/* spacing between data		*/
	  unsigned long	last_disp ;	/* last time displayed		*/
	  Terrain_data	*data ;		/* pointer to data, y varies fastest */
	  Desc_Node	*child ;
	  char		*name ;		/* terrain's filename		*/
	  float		max_disp ;	/* viewing threshold**2		*/
	  struct terrain_node *next ;	/* next terrain in root list	*/
	} ;


struct transform_node {
	  Object_Type	type ;		/* = TRANSFORM_NODE		*/
	  int		(*func)() ;	/* draw_transform_node()	*/
	  Matrix3df	matrix ;
	  Matrix3df	imatrix ;
	  Desc_Node	*child ;
	} ;


struct vasi_node {
	  Object_Type	type ;		/* = VASI_NODE		*/
	  int		(*func)() ;	/* draw_vasi_node()	*/
	  int		vasi_type ;	/* 0=2-bar */
	  Point3df	l1,v1 ;		/* first light */
	  Point3df	l2,v2 ;		/* second light */
	} ;



struct beacon_node {
	  Object_Type	type ;		/* = BEACON_NODE	*/
	  int		(*func)() ;	/* draw_beacon_node()	*/
	  int		beacon_type ;	/* 0=civilian */
	  Point3df	posn ;		/* position */
	  int		state,count ;	/* light state & count */
	} ;

#endif	OBJECT_TYPES
