#ifndef lint
static	char sccsid[] = "@(#)terrain.c 1.1 89/03/10 SMI";
#endif



/* This file performs general terrain management.
 *
 * There are several lists:
 *
 *   terrain objects: all *.terrain files that were found at
 *	start-up time.  Each object contains its location (long, lat),
 *	its four corners (x,y,z,c), its color, its shading method,
 *	its size, its grid size, its grid dimension, the last time
 *	it was actually drawn, a pointer to the actual terrain
 *	data (if loaded), a pointer to the objects in that terrain
 *	(if loaded), the filename that it was found in, the filename
 *	that the objects were in, the maximum display distance,
 *
 * Elevation data is stored in the file by rows.  It is stored in memory
 * by columns.
 *
 *
 * Shading:
 *   When terrain objects are read in from file, and once every few minutes,
 *   a color is assigned to each square based on the object's color, the
 *   square's angle to the sun, the distance from the plane, visibility.  If
 *   the nearest corner of the terrain or patch is greater than 1/2
 *   "visibility", it is shaded with "fog" colors (see colors.c).
 *   Otherwise, the object is shaded with it's requested colors.
 *
 *
 *
 * Drawing:
 *   terrain objects are not drawn if their distance from the plane is
 *   greater than "visibility".  Terrain objects are drawn as a single
 *   square if the distance to the nearest corner is greater than the
 *   maximum display distance.  The maximum display distance is
 *   approximately the distance at which the terrain occupies fifteen
 *   degrees of view.  An object will be loaded from file whenever
 *   it comes within the maximum display distance for the first time.
 *
 *
 * Bookkeeping:
 *   An object which hasn't been displayed for five minutes will be
 *   removed.
 *
 */

#define	LINELEN	256
#define	FILELEN	40



#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/dir.h>
#include "gfx_mac.h"
#include "flight.h"
#include "object_types.h"
#include "graphics.h"





static	Terrain_Node	*first_terrain ;


extern	Desc_Node	bay_object ;

extern	int		current_color ;

#ifdef	DEBUG
#define	register
#endif	DEBUG


/* shade(color, v1,v2)
 *
 * shade a patch or square based on the normal vector derived from
 * vectors v1 & v2.  It is known that v1.y and v2.x == 0.0
 */

static	int
shade(color, v1,v2)
	int		color ;
register Point3df	*v1,*v2 ;
{
	Point3df	normal ;
register float		a ;
register int		c0,c1 ;

/*
 *	normal.x = v1->y*v2->z - v1->z*v2->y ;
 *	normal.y = v1->z*v2->x - v1->x*v2->z ;
 *	normal.z = v1->x*v2->y - v1->y*v2->x ;
 */

	normal.x = - v1->z*v2->y ;
	normal.y = - v1->x*v2->z ;
	normal.z = v1->x*v2->y ;

	normal.z *= .2 ;
	a = normal.x*normal.x + normal.y*normal.y + normal.z*normal.z ;
	if( a > 0.0 )
	{
	  a = sqrt(a) ;
	  a = 1.0/a ;
	  normal.x *= a ;
	  normal.y *= a ;
	  normal.z *= a ;
	}

	a = normal.x*Sun_vec.x + normal.y*Sun_vec.y + normal.z*Sun_vec.z ;
	if( a < 0.0 )
	  a = 0.0 ;

	c0 = Color_Mapping[color][0] ;
	c1 = Color_Mapping[color][1] ;

	c0 += a * (c1-c0) ;
	if( !hwdb )
	  c0 = Gfx_dbuf_get_color(Main_gfx,c0) ;
	return c0 ;
}




/*
 * Initialize terrain list, search for *.terrain files in the current
 * directory.
 */

Init_Terrain()
{
	DIR		*dir ;
	struct direct	*entry ;
	int		i,j,n ;
	FILE		*ifile ;
	Point3df	v1,v2 ;		/* vectors for shading */


	first_terrain = NULL ;

	dir = opendir(".") ;
	while( (entry = readdir(dir)) != NULL )
	{
	  i = entry->d_namlen - 8 ;
	  if( i > 0  &&  strcmp( &entry->d_name[i], ".terrain" ) == 0 )
	  {
	    ifile = fopen(entry->d_name, "r") ;
	    if( ifile != NULL )
	    {
	      register Terrain_Node *node ;
	      char line[LINELEN] ;
	      char *str ;

	      node = (Terrain_Node *) malloc(sizeof(Terrain_Node)) ;
	      node->next = first_terrain ;
	      first_terrain = node ;

	      node->name = (char *) malloc(entry->d_namlen+1) ;
	      bcopy( entry->d_name, node->name, entry->d_namlen+1 ) ;

	      /* read & discard comments */

	      do {
		str = fgets(line, LINELEN, ifile) ;
	      }
	      while( str != NULL  &&  (line[0] == ';' || line[0] == '\n' )) ;

	      /* gouraud shading doesn't work with color table d.b. */
	      if( hwdb  &&  strncmp(line,"gouraud",7) == 0 )
	      {
		node->type = G_TERRAIN_NODE ;
		node->func = draw_g_terrain_node ;
	      }
	      else
	      {
		node->type = TERRAIN_NODE ;
		node->func = draw_terrain_node ;
	      }

	      fgets(line, LINELEN, ifile) ;
	      node->color = strncmp(line,"green",5) ? CITY : TERRAIN ;

	      n = fscanf(ifile, "%f %f", &node->x0, &node->y0) ;
	      n = fscanf(ifile, "%f %f", &node->sx, &node->sy) ;
	      n = fscanf(ifile, "%d %d", &node->nx, &node->ny) ;

	      node->p0.p.x = node->x0 ;
	      node->p0.p.y = node->y0 ;
	      node->p1.p.x = node->x0 + node->sx ;
	      node->p1.p.y = node->y0 ;
	      node->p2.p.x = node->x0 + node->sx ;
	      node->p2.p.y = node->y0 + node->sy ;
	      node->p3.p.x = node->x0 ;
	      node->p3.p.y = node->y0 + node->sy ;
	      node->max_ht = 0.0 ;

	      node->dx = node->sx / (node->nx-1) ;
	      node->dy = node->sy / (node->ny-1) ;

	      node->last_disp = 0 ;
	      node->data = NULL ;
	      node->child = NULL ;

	      if( node->sx > node->sy )
		node->max_disp = node->sx ;
	      else
		node->max_disp = node->sy ;
	      node->max_disp *= 4 ;		/* ~= 1/sin(15) */
	      node->max_disp *= node->max_disp ;	/* square it */

	      /* scan through the data for elevations */

	      for(i=0; i < node->ny; ++i)
		for(j=0; j < node->nx; ++j)
		{
		  int elev ;
		  n = fscanf(ifile, "%d", &elev) ;
		  if( elev > node->max_ht )
		    node->max_ht = elev ;
		  if( i == 0 )
		  {
		    if( j == 0 )
		      node->p0.p.z = elev ;
		    else if( j == node->nx - 1 )
		      node->p1.p.z = elev ;
		  }
		  else if( i == node->ny - 1 )
		  {
		    if( j == 0 )
		      node->p3.p.z = elev ;
		    else if( j == node->nx - 1 )
		      node->p2.p.z = elev ;
		  }
		}

	      /* shade the patch */

	      v1.x = node->sx ;
	      v1.z = node->p1.p.z - node->p0.p.z ;
	      v2.y = node->sy ;
	      v2.z = node->p3.p.z - node->p0.p.z ;
	      node->p0.color = shade(node->color, &v1, &v2) ;

	      /* kludge until we have object loading: */

	      if( strcmp(node->name,"pao.terrain") == 0 )
		node->child = &bay_object ;

	      fclose(ifile) ;
	    }
	  }
	}
	closedir(dir) ;


}



/*
 * shade all the squares in a terrain patch
 */

Shade_terrain(node)
register Terrain_Node	*node ;
{
	int		i,j,n ;
	Point3df	v1,v2 ;

	v1.x = node->sx ;
	v1.y = 0.0 ;
	v1.z = node->p1.p.z - node->p0.p.z ;
	v2.x = 0.0 ;
	v2.y = node->sy ;
	v2.z = node->p3.p.z - node->p0.p.z ;
	node->p0.color = shade(node->color, &v1, &v2) ;

	if( node->data != NULL )
	{
	  Terrain_data *ptr = node->data ;
	  register Terrain_data *dp0, *dp1, *dp2 ;

	  v1.x = node->dx ;
	  v2.y = node->dy ;

	  for( i=0; i < node->nx-1; ++i )
	  {
	    dp0 = dp1 = dp2 = node->data + i*node->ny ;
	    dp1 += node->ny ;
	    ++dp2 ;
	    for( j=0; j < node->ny-1; ++j )
	    {
	      v1.z = dp1->elev - dp0->elev ;
	      v2.z = dp2->elev - dp0->elev ;
	      dp0->color = shade(node->color, &v1, &v2) ;
	      ++dp0 ;
	      ++dp1 ;
	      ++dp2 ;
	    }
	  }
	  if( node->type == G_TERRAIN_NODE )	/* needs a little bit more */
	  {
	    v1.x = node->dx ;
	    v2.y = node->dy ;
	    dp0 = node->data + node->ny - 1 ;
	    dp1 = dp0 + node->ny ;
	    dp2 = dp0 - 1 ;
	    for( i=1; i < node->nx; ++i )	/* fill in max-y profile */
	    {
	      v1.z = dp1->elev - dp0->elev ;
	      v2.z = dp0->elev - dp2->elev ;
	      dp0->color = shade(node->color, &v1,&v2) ;
	      dp0 += node->ny ;
	      dp1 += node->ny ;
	      dp2 += node->ny ;
	    }
	    v1.x = node->dx ;
	    v2.y = node->dy ;
	    dp0 = node->data + (node->nx-1)*node->ny ;
	    dp1 = dp0 - node->ny ;
	    dp2 = dp0 + 1 ;
	    for( j=1; j < node->ny; ++j )	/* fill in max-x profile */
	    {
	      v1.z = dp0->elev - dp1->elev ;
	      v2.z = dp2->elev - dp0->elev ;
	      dp0->color = shade(node->color, &v1,&v2) ;
	      ++dp0 ;
	      ++dp1 ;
	      ++dp2 ;
	    }
	    					/* fill in corner element */
	    v1.x = node->dx ;
	    v2.y = node->dy ;
	    dp0 = node->data + node->ny*node->nx - 1 ;
	    dp1 = dp0 - node->ny ;
	    dp2 = dp0 - 1 ;
	    v1.z = dp0->elev - dp1->elev ;
	    v2.z = dp0->elev - dp2->elev ;
	    dp0->color = shade(node->color, &v1,&v2) ;
	  }
	}
}



/*
 * load terrain data from a file
 */

Load_terrain(node)
register Terrain_Node	*node ;
{
	int		i,j,n ;
	FILE		*ifile ;


	ifile = fopen(node->name, "r") ;
	if( ifile != NULL )
	{
	  char line[LINELEN] ;
	  char *str ;
	  Terrain_data *ptr ;

	  node->data = ptr = (Terrain_data *)
		malloc(node->nx * node->ny * sizeof(Terrain_data) ) ;

	  /* read & discard comments */

	  do {
	    str = fgets(line, LINELEN, ifile) ;
	  }
	  while( str != NULL  &&  (line[0] == ';' || line[0] == '\n' )) ;

	  fgets(line, LINELEN, ifile) ;		/* discard color line too */

	  /* re-read position, size and dimension */

	  n = fscanf(ifile, "%f %f", &node->x0, &node->y0) ;
	  n = fscanf(ifile, "%f %f", &node->sx, &node->sy) ;
	  n = fscanf(ifile, "%d %d", &node->nx, &node->ny) ;

	  /* scan through the data for elevations */

	  for(i=0; i < node->ny; ++i)
	  {
	    register Terrain_data *dp0 ;
	    dp0 = ptr ;
	    for(j=0; j < node->nx; ++j)
	    {
	      n = fscanf(ifile, "%hd", &dp0->elev) ;
	      dp0 += node->ny ;
	    }
	    ++ptr ;
	  }
	  fclose(ifile) ;

	  Shade_terrain(node) ;
	}
}


/*
 * go through list of terrain nodes, reshade everything.
 */

Reshade_Terrain()
{
	Terrain_Node	*node = first_terrain ;

	while( node != NULL )
	{
	  Shade_terrain(node) ;
	  node = node->next ;
	}
}




/*
 * go through list of terrain nodes, unloading anything that's more than
 * five minutes old.
 */

GC_Terrain()
{
	Terrain_Node	*node = first_terrain ;

	while( node != NULL )
	{
	  if( Now.tv_sec - node->last_disp > 300 )
	  {
	    free( node->data ) ;
	    node->data = NULL ;

	    /* free up children, do this later */
	  }
	  node = node->next ;
	}
}



/*
 * top level terrain handling, scan through all terrain objects
 * and pass to draw_terrain_node()
 */

Draw_Terrain(viewpt, matrix)
	Point3df	*viewpt ;		/* inverse xform of Me_Posn */
	Matrix3df	matrix ;		/* view matrix */
{
	Terrain_Node	*node = first_terrain ;

	current_color = -1 ;
	while( node != NULL )
	{
	  (*node->func)(node, viewpt, matrix) ;
	  node = node->next ;
	}
}




dump_terrain(node)
	Terrain_Node	*node ;
{
	Terrain_data	*data = node->data ;
	int	i,j ;
	Terrain_data	*ptr ;

	for(j=0; j<node->ny; ++j)
	{
	  for(i=0; i<node->nx; ++i)
	  {
	    ptr = data + i*node->ny + j ;
	    printf("%4d:%2d",ptr->elev,ptr->color) ;
	  }
	  putchar('\n') ;
	}
}
